# import Asclepius dependencies
from asclepius.instelling import GGZ, ZKH, HardCodedParameters

# import other dependencies
from pandas import read_excel, merge, isnull, DataFrame
from typing import Union

class TestFuncties:

    def __init__(self):
        pass

        # DAILY AUDIT FUNCTIES

    def wrangle_da(self, excel_a, excel_p) -> DataFrame:
        # import excelsheets
        data_a = read_excel(excel_a, index_col = None, header = 2)
        data_p = read_excel(excel_p, index_col = None, header = 2)

        if data_a.columns.values[0] == 'Controle/norm':
            # rename columns
            data_a = data_a.rename(columns = {'Controle/norm': 'controle', 'Aantal': 'aantal_a', 'Impact': 'impact_a'})
            data_p = data_p.rename(columns = {'Controle/norm': 'controle', 'Aantal': 'aantal_p', 'Impact': 'impact_p'})
        else:
            # rename columns
            data_a = data_a.rename(columns = {'Controle': 'controle', 'Aantal': 'aantal_a', 'Impact': 'impact_a'})
            data_p = data_p.rename(columns = {'Controle': 'controle', 'Aantal': 'aantal_p', 'Impact': 'impact_p'})
            
        # Outer join van de dataframes op controle
        wrangled_data = merge(data_a, data_p, how = 'outer', on = 'controle')

        # drop impact_a and impact_p
        wrangled_data = wrangled_data.drop(columns=['impact_a', 'impact_p'])

        # fill NaN values
        wrangled_data['aantal_a'] = wrangled_data['aantal_a'].fillna(0)

        # add absolute and percentual difference
        wrangled_data = wrangled_data.assign(diff_abs = wrangled_data['aantal_a'] - wrangled_data['aantal_p'])
        wrangled_data = wrangled_data.assign(diff_pct = round((wrangled_data['diff_abs'] / wrangled_data['aantal_p']) * 100, 2))

        # if P is NaN
        wrangled_data['aantal_p'] = wrangled_data['aantal_p'].fillna(0)
        wrangled_data['diff_abs'] = wrangled_data['diff_abs'].fillna(wrangled_data['aantal_a'])
        wrangled_data['diff_pct'] = wrangled_data['diff_pct'].fillna(100.00)

        return wrangled_data

    def check_verschillen_da(self, wrangled_data: DataFrame, tolerantie_abs: int, tolerantie_pct: int):
        verschillen = DataFrame(columns = list(wrangled_data.columns.values))
        for i in range(len(wrangled_data)):
            entry = wrangled_data.iloc[i, :]
            if abs(entry['diff_abs']) > tolerantie_abs and abs(entry['diff_pct']) > tolerantie_pct:
                verschillen = verschillen.append(entry)
        return verschillen

    def aantallencheck(self, instelling: Union[GGZ, ZKH], test: bool = False):
        if test:
            excel_a = instelling.excel_da_a_test
            excel_p = instelling.excel_da_p_test
        else:
            excel_a = instelling.excel_da_a
            excel_p = instelling.excel_da_p
        
        actieaantallen = self.wrangle_da(excel_a, excel_p)
        bevindingen = self.check_verschillen_da(actieaantallen, instelling.tolerantie_abs, instelling.tolerantie_pct)
        if len(bevindingen) == 0:
            print('Geen significante verschillen gevonden')
        else:
            pass

        if test:
            instelling.bevindingen_da_test = bevindingen
        else:
            instelling.bevindingen_da = bevindingen
        return None

    # PRESTATIEKAART FUNCTIES

    def wrangle_pk(self, excel_a, excel_p):
        # import excelsheets
        data_a = read_excel(excel_a, index_col = None, header = 0)
        data_p = read_excel(excel_p, index_col = None, header = 0)

        # drop Verschil column
        data_a = data_a.drop(columns=['Verschil'])
        data_p = data_p.drop(columns=['Verschil'])

        # voeg index column toe
        data_a['index_a'] = data_a.index
        data_p['index_p'] = data_p.index

        # hernoem kolommen
        data_a = data_a.rename(columns = {'Titel': 'titel', 'Norm': 'norm_a', 'Realisatie': 'real_a', 'sectie': 'sectie_a'})
        data_p = data_p.rename(columns = {'Titel': 'titel', 'Norm': 'norm_p', 'Realisatie': 'real_p', 'sectie': 'sectie_p'})

        # join dataframes op titlel
        wrangled_data = merge(data_a, data_p, how = 'outer', on = 'titel')

        # return de prestatiekaart data
        return wrangled_data

    def check_existence(self, prestatiekaart, bevindingen):
        # checkt of een titel niet in A bestaat, niet in P bestaat
        for i in range(len(prestatiekaart)):
            if isnull(prestatiekaart['index_a'][i]):
                new_row = {'Indicator': prestatiekaart['titel'][i], 'Norm/Realisatie': 'Beiden', 'A': '', 'P': '', 'Bevinding': "Indicator niet in A portaal."}
                bevindingen = bevindingen.append(new_row, ignore_index = True)
                prestatiekaart = prestatiekaart.drop(index = i)
            elif isnull(prestatiekaart['index_p'][i]):
                new_row = {'Indicator': prestatiekaart['titel'][i], 'Norm/Realisatie': 'Beiden', 'A': '', 'P': '', 'Bevinding': "Indicator niet in P portaal."}
                bevindingen = bevindingen.append(new_row, ignore_index = True)
                prestatiekaart = prestatiekaart.drop(index = i)
            else:
                pass
        prestatiekaart = prestatiekaart.reset_index(drop = True)
        
        return prestatiekaart, bevindingen

    def check_empty(self, prestatiekaart, bevindingen):
        # checkt of een titel niet leeg is in A en P
        for i in range(len(prestatiekaart)):
            if isnull(prestatiekaart['norm_a'][i]) and isnull(prestatiekaart['real_a'][i]) and isnull(prestatiekaart['norm_p'][i]) and isnull(prestatiekaart['real_p'][i]):
                new_row = {'Indicator': prestatiekaart['titel'][i], 'Norm/Realisatie': 'Beiden', 'A': '', 'P': '', 'Bevinding': "Indicator is leeg in beide portalen."}
                bevindingen = bevindingen.append(new_row, ignore_index = True)
                prestatiekaart = prestatiekaart.drop(index = i)
            elif isnull(prestatiekaart['norm_a'][i]) and isnull(prestatiekaart['real_a'][i]):
                new_row = {'Indicator': prestatiekaart['titel'][i], 'Norm/Realisatie': 'Beiden', 'A': '', 'P': '', 'Bevinding': "Indicator is leeg in A portaal."}
                bevindingen = bevindingen.append(new_row, ignore_index = True)
                prestatiekaart = prestatiekaart.drop(index = i)
            elif isnull(prestatiekaart['norm_p'][i]) and isnull(prestatiekaart['real_p'][i]):
                new_row = {'Indicator': prestatiekaart['titel'][i], 'Norm/Realisatie': 'Beiden', 'A': '', 'P': '', 'Bevinding': "Indicator is leeg in P portaal."}
                bevindingen = bevindingen.append(new_row, ignore_index = True)
                prestatiekaart = prestatiekaart.drop(index = i)
            else:
                pass
        prestatiekaart = prestatiekaart.reset_index(drop = True)
        
        return prestatiekaart, bevindingen

    def check_normen(self, prestatiekaart, bevindingen):
        for i in range(len(prestatiekaart)):
            # Als beide normen leeg zijn zal daar waarschijnlijk een goede reden voor zijn
            if isnull(prestatiekaart['norm_a'][i]) and isnull(prestatiekaart['norm_p'][i]):
                pass
            # check of de norm niet leeg is in het A portaal
            elif isnull(prestatiekaart['norm_a'][i]):
                new_row = {'Indicator': prestatiekaart['titel'][i], 'Norm/Realisatie': 'Norm', 'A': '', 'P': '', 'Bevinding': "Norm leeg in A portaal."}
                bevindingen = bevindingen.append(new_row, ignore_index = True)
            # check of de norm niet leeg is in het P portaal
            elif isnull(prestatiekaart['norm_p'][i]):
                new_row = {'Indicator': prestatiekaart['titel'][i], 'Norm/Realisatie': 'Norm', 'A': '', 'P': '', 'Bevinding': "Norm leeg in P portaal."}
                bevindingen = bevindingen.append(new_row, ignore_index = True)
            # check of de normen in het A en het P portaal gelijk zijn aan elkaar
            elif not (prestatiekaart['norm_a'][i] == prestatiekaart['norm_p'][i]):
                new_row = {'Indicator': prestatiekaart['titel'][i], 'Norm/Realisatie': 'Norm', 'A': str(prestatiekaart['norm_a'][i]), 'P': str(prestatiekaart['norm_p'][i]), 'Bevinding': 'Norm in A wijkt af van P.'}
                bevindingen = bevindingen.append(new_row, ignore_index = True)
            else:
                pass
        
        # geef de prestatiekaart en bevindingen als output
        return prestatiekaart, bevindingen

    def check_realisaties(self, prestatiekaart, bevindingen):
        for i in range(len(prestatiekaart)):
            if isnull(prestatiekaart['real_a'][i]) or isnull(prestatiekaart['real_p'][i]):
                if isnull(prestatiekaart['real_a'][i]) and isnull(prestatiekaart['real_p'][i]):
                    new_row = {'Indicator': prestatiekaart['titel'][i], 'Norm/Realisatie': 'Realisatie', 'A': '', 'P': '', 'Bevinding': "Realisatie leeg in A en P portaal."}
                    bevindingen = bevindingen.append(new_row, ignore_index = True)
                else:
                    if isnull(prestatiekaart['real_a'][i]):
                        new_row = {'Indicator': prestatiekaart['titel'][i], 'Norm/Realisatie': 'Realisatie', 'A': '', 'P': '', 'Bevinding': "Realisatie leeg in A portaal."}
                        bevindingen = bevindingen.append(new_row, ignore_index = True)
                    else:
                        pass
                    if isnull(prestatiekaart['real_p'][i]):
                        new_row = {'Indicator': prestatiekaart['titel'][i], 'Norm/Realisatie': 'Realisatie', 'A': '', 'P': '', 'Bevinding': "Realisatie leeg in P portaal."}
                        bevindingen = bevindingen.append(new_row, ignore_index = True)
                    else:
                        pass
            elif not (prestatiekaart['real_a'][i] == prestatiekaart['real_p'][i]):
                new_row = {'Indicator': prestatiekaart['titel'][i], 'Norm/Realisatie': 'Realisatie', 'A': str(prestatiekaart['real_a'][i]), 'P': str(prestatiekaart['real_p'][i]), 'Bevinding': 'Realisatie in A wijkt af van P.'}
                bevindingen = bevindingen.append(new_row, ignore_index = True)
            else:
                pass
            
        return prestatiekaart, bevindingen

    def aantal_indicatoren(self, wrangled_data: DataFrame):
        lengte_a = wrangled_data['index_a'].max() + 1
        lengte_p = wrangled_data['index_p'].max() + 1
        return lengte_a, lengte_p

    def conclusie(self, bevindingen, lengte_a: int, lengte_p: int):
        lengte_bevingingen = len(bevindingen)
        new_row = {'Indicator': 'Aantal indicatoren in Prestatiekaart:', 'Norm/Realisatie': '', 'A': str(lengte_a), 'P': str(lengte_p), 'Bevinding': ''}
        bevindingen = bevindingen.append(new_row, ignore_index = True)
        new_row = {'Indicator': 'Totaal aantal bevindingen:', 'Norm/Realisatie': '', 'A': '', 'P': '', 'Bevinding': lengte_bevingingen}
        bevindingen = bevindingen.append(new_row, ignore_index = True)
        return bevindingen
    
    def prestatiekaarten_vergelijken(self, instelling: Union[GGZ, ZKH], product: str):
        if product == 'bi':
            excel_a = instelling.excel_bi_a
            excel_p = instelling.excel_bi_p
        elif product == 'zpm':
            excel_a = instelling.excel_zpm_a
            excel_p = instelling.excel_zpm_p
        elif product == 'zpm_nza':
            excel_a = instelling.excel_zpm_nza_a
            excel_p = instelling.excel_zpm_nza_p
        else:
            print('Geen valide prestatiekaart om te vergelijken!')
            pass

        prestatiekaart = self.wrangle_pk(excel_a, excel_p)
        bevindingen = DataFrame({'Indicator': [], 'Norm/Realisatie': [], 'A': [], 'P': [], 'Bevinding': []})

        lengte_a, lengte_p = self.aantal_indicatoren(prestatiekaart)
        
        prestatiekaart, bevindingen = self.check_existence(prestatiekaart, bevindingen)
        
        prestatiekaart, bevindingen = self.check_empty(prestatiekaart, bevindingen)
        
        prestatiekaart, bevindingen = self.check_normen(prestatiekaart, bevindingen)
        
        prestatiekaart, bevindingen = self.check_realisaties(prestatiekaart, bevindingen)

        bevindingen = self.conclusie(bevindingen, lengte_a, lengte_p)
        
        if product == 'bi':
            instelling.bevindingen_bi = bevindingen
        elif product == 'zpm':
            instelling.bevindingen_zpm = bevindingen
        elif product == 'zpm_nza':
            instelling.bevindingen_zpm_nza = bevindingen
        else:
            pass
        return None



class Verklaren:

    def __init__(self) :
        pass

    def splits_controle_omschrijving(self, bevindingen: DataFrame) -> DataFrame:
        gesplitste_controle = bevindingen["controle"].str.split(pat=" - ", n=1, expand = True)
        #bevindingen = bevindingen.assign(kenmerk = gesplitste_controle[0], omschrijving = gesplitste_controle[1])
        bevindingen['kenmerk'] = gesplitste_controle[0].to_list()
        bevindingen['omschrijving'] = gesplitste_controle[1].to_list()
        bevindingen = bevindingen[['kenmerk', 'omschrijving', 'aantal_a', 'aantal_p', 'diff_abs', 'diff_pct']]
        return bevindingen

    def standaardverschillen_da(self, instelling: Union[GGZ, ZKH], test: bool = False):
        # Check of de instelling in het standaardverschillen dict voorkomt
        if instelling.klant_code in HardCodedParameters.standaardverschillen_da:
            if test:
                bevindingen = instelling.bevindingen_da_test
            
            else:
                bevindingen = instelling.bevindingen_da
            
            if len(bevindingen) == 0:
                pass
            else:
                bevindingen = self.splits_controle_omschrijving(bevindingen)
                bevindingen['verklaring'] = ''

                # Check of de controle/norm in de standaardverschillen van de instelling voorkomt
                for i in range(len(bevindingen)):
                    entry = bevindingen.iloc[i, :]

                    # Als de controle/norm hierin voorkomt krijgt deze de verklaring 'standaardverschil'
                    if entry['kenmerk'] in set(HardCodedParameters.standaardverschillen_da[instelling.klant_code]):
                        bevindingen.iloc[i,-1] = "Standaardverschil"
                    else:
                        pass
            
            # Vervang het bevindingen_da dataframe voor die met de standaardverschillen
            if test:
                instelling.bevindingen_da_test = bevindingen
            
            else:
                instelling.bevindingen_da = bevindingen
        else:
            pass
        return None
    
    


# import hardcoded information from fileserver
import sys
sys.path.append(r'\\fileserver.valuecare.local\Algemeen\Automatisch testen\Python')
from parameters import HardCodedParameters

# import other dependencies
from pandas import DataFrame

class Instelling:
    def __init__(self, klant_code: str):
        self.klant_code = klant_code
        self.baselink = ''
        self.huidige_omgeving = ''
        self.link_dict = {}

        # Daily Audit
        self.tolerantie_abs = 0
        self.tolerantie_pct = 0
        self.excel_da_a = 'excel_da_a file'
        self.excel_da_p = 'excel_da_p file'
        self.bevindingen_da = DataFrame()

        self.excel_da_a_test = 'excel_da_test_a file'
        self.excel_da_p_test = 'excel_da_test_p file'
        self.bevindingen_da_test = DataFrame()

        # BI prestatiekaart
        self.excel_bi_a = 'excel_bi_a file'
        self.excel_bi_p = 'excel_bi_p file'
        self.bevindingen_bi = DataFrame()

        # ZPM prestatiekaart
        self.excel_zpm_a = 'excel_zpm_a file'
        self.excel_zpm_p = 'excel_zpm_p file'
        self.bevindingen_zpm = DataFrame()

        # ZPM_NZA prestatiekaart
        self.excel_zpm_nza_a = 'excel_zpm_a file'
        self.excel_zpm_nza_p = 'excel_zpm_p file'
        self.bevindingen_zpm_nza = DataFrame()

        # SLM deltas totaal
        self.slm_delta_a = ''
        self.slm_delta_p = ''
        return None
    
    def update_baselink(self):
        # login link
        self.login = self.baselink + self.link_dict['login']
        
        # DA links
        self.daily_audit = self.baselink + self.link_dict['daily_audit']
        self.da_excel_download = self.daily_audit + self.link_dict['da_excel_download']
        self.daily_audit_test = self.baselink + self.link_dict['daily_audit_test']
        self.da_test_excel_download = self.daily_audit_test + self.link_dict['da_test_excel_download']
        self.da_excel_groot_download = self.daily_audit + self.link_dict['da_excel_groot_download']

        # BI links
        self.bi_prestatiekaart = self.baselink + self.link_dict['bi_prestatiekaart']
        self.bi_excel_download = self.baselink + self.link_dict['bi_excel_download']

        # zpm links
        self.zpm_prestatiekaart = self.baselink + self.link_dict['zpm_prestatiekaart']
        self.zpm_excel_download = self.baselink + self.link_dict['zpm_excel_download']

        # zpm_nza links
        self.zpm_nza_prestatiekaart = self.baselink + self.link_dict['zpm_nza_prestatiekaart']
        self.zpm_nza_excel_download = self.baselink + self.link_dict['zpm_nza_excel_download']

        # slm links
        self.slm_per_verzekeraar = self.baselink + self.link_dict['slm_per_verzekeraar']

        return None

    def kies_omgeving(self, omgeving: str):
        if omgeving == 'acceptatie':
            # verander huidige omgeving naar acceptatie
            self.huidige_omgeving = 'acceptatie'
        elif omgeving == 'productie':
            # verander huidige omgeving naar productie
            self.huidige_omgeving = 'productie'
        else:
            pass
        
        self.baselink = HardCodedParameters.baselink + self.huidige_omgeving[0] + self.klant_code + '/portaal/'
        self.update_baselink()
        return None
    
    def genereer_nieuwe_naam(self, product: str, test: bool = False):
        if test:
            new_name = f'{self.klant_code}_{product}_{self.huidige_omgeving[0]}_test.xlsx'
        else:
            new_name = f'{self.klant_code}_{product}_{self.huidige_omgeving[0]}.xlsx'
        return new_name
    
    def update_bestand_locatie(self, product, new_path, test):
        if test:
            attr = f'excel_{product}_{self.huidige_omgeving[0]}_test'
        else:
            attr = f'excel_{product}_{self.huidige_omgeving[0]}'
        
        setattr(self, attr, new_path)
        return None
    
    def set_slm_delta(self, rode_delta: str):
        attr = f'slm_delta_{self.huidige_omgeving[0]}'
        setattr(self, attr, rode_delta)
        return None


class ZKH(Instelling):

    def __init__(self, klant_code: str):
        super().__init__(klant_code)

        self.da = True
        self.tolerantie_abs = 7 # 7
        self.tolerantie_pct = 2 # 2

        self.link_dict = HardCodedParameters.zkh_dict

        self.update_baselink()
        return None


class GGZ(Instelling):

    variabele_links = HardCodedParameters.ggz_variabele_links

    def __init__(self, klant_code: str):
        super().__init__(klant_code)

        self.link_dict = HardCodedParameters.ggz_dict

        self.update_link_dict()

        self.update_baselink()
        return None
    
    def update_link_dict(self):
        if self.klant_code in GGZ.variabele_links.index.tolist():
            self.link_dict['bi_prestatiekaart'] = GGZ.variabele_links['bi_prestatiekaart'][self.klant_code]
            self.link_dict['bi_excel_download'] = GGZ.variabele_links['bi_excel_download'][self.klant_code]
        else:
            pass
        return None


# import Asclepius dependencies
from asclepius.instelling import GGZ, ZKH
from asclepius.medewerker import Medewerker

# import other dependencies
from selenium import webdriver
from typing import Union
from time import sleep

class PortaalDriver:
    driver = 'C:\Program Files (x86)\chromedriver.exe'
    
    def __init__(self, gebruiker: Medewerker):
        self.gebruiker = gebruiker
        self.driver = PortaalDriver.driver
        self.portaal = None
        return None
    
    def open_portaal(self):
        self.portaal = webdriver.Chrome(self.driver)
        return None
    
    def inloggen(self, instelling: Union[GGZ, ZKH]):
        # open portaal
        self.portaal = webdriver.Chrome(self.driver)
        
        # haal de loginpagina op
        self.portaal.get(instelling.login)
        sleep(2)
        
        # vind invoer op de pagina
        username = self.portaal.find_element_by_name('username')
        password = self.portaal.find_element_by_name('password')
        submit = self.portaal.find_element_by_name('submit')
        
        # verstuur de informatie van de gebruiker
        username.send_keys(self.gebruiker.gebruikersnaam)
        password.send_keys(self.gebruiker.wachtwoord)
        submit.click()
        sleep(2)
        return

    # Download Excel Functies

    def download_da_excel(self, instelling: Union[GGZ, ZKH], test: bool = False):
        if test:
            da_link = instelling.daily_audit_test
            da_download_link = instelling.da_test_excel_download
        else:
            da_link = instelling.daily_audit
            da_download_link = instelling.da_excel_download
        
        self.portaal.get(da_link)
        sleep(2)
        if (not test) and (instelling.huidige_omgeving == "productie"):
            quickremove = self.portaal.find_element_by_class_name('quickRemove')
            quickremove_list = quickremove.find_elements_by_tag_name("li")
            for quickremove_item in quickremove_list:
                if quickremove_item.text == " Behandeld: Ja" or quickremove_item.text == " Behandeld: Nee":
                    delete_behandeld = quickremove_item.find_element_by_class_name('delete')
                    delete_behandeld.click()
                    sleep(3)
                    break
        self.portaal.get(da_download_link)
        sleep(5)
        return None

    def download_bi_excel(self, instelling: Union[GGZ, ZKH]):
        self.portaal.get(instelling.bi_prestatiekaart)
        sleep(2)
        self.portaal.get(instelling.bi_excel_download)
        sleep(5)
        return None

    def download_zpm_excel(self, instelling: Union[GGZ, ZKH]):
        self.portaal.get(instelling.zpm_prestatiekaart)
        sleep(2)
        self.portaal.get(instelling.zpm_excel_download)
        sleep(5)
        return None
    
    def download_zpm_nza_excel(self, instelling: Union[GGZ, ZKH]):
        self.portaal.get(instelling.zpm_nza_prestatiekaart)
        sleep(2)
        self.portaal.get(instelling.zpm_nza_excel_download)
        sleep(5)
        return None
    
    def get_slm_delta(self, instelling: Union[GGZ, ZKH]):
        self.portaal.get(instelling.slm_per_verzekeraar)
        sleep(2)
        totaal_footer = self.portaal.find_element_by_tag_name('tfoot')
        totaal_list = totaal_footer.find_elements_by_tag_name('th')
        rode_delta = totaal_list[-1].text
        return rode_delta

    # Webscraper Portaal Functies

    def webscraper_da_portaal(self, instelling: Union[GGZ, ZKH]):
        # inloggen op het portaal
        self.inloggen(instelling)

        # download excel controle/norm in productie
        self.download_da_excel(instelling, False)
        self.gebruiker.webscraper_hernoem_bestand(instelling, 'da', False)
        
        sleep(1)
        
            # download excel controle/norm in test
        self.download_da_excel(instelling, True)
        self.gebruiker.webscraper_hernoem_bestand(instelling, 'da', True)

        # sluit het portaal af
        self.portaal.close()
        return None

    def webscraper_bi_portaal(self, instelling: Union[GGZ, ZKH]):
        # inloggen op het portaal
        self.inloggen(instelling)
        
            # download excel BI prestatiekaart
        self.download_bi_excel(instelling)
        self.gebruiker.webscraper_hernoem_bestand(instelling, 'bi')
        

        # sluit het portaal af
        self.portaal.close()
        return None

    def webscraper_zpm_portaal(self, instelling: Union[GGZ, ZKH]):
        # inloggen op het portaal
        self.inloggen(instelling)

        # download excel ZPM prestatiekaart
        self.download_zpm_excel(instelling)
        self.gebruiker.webscraper_hernoem_bestand(instelling, 'zpm')

        # sluit het portaal af
        self.portaal.close()
        return None

    def webscraper_zpm_nza_portaal(self, instelling: Union[GGZ, ZKH]):
        # inloggen op het portaal
        self.inloggen(instelling)

        # download excel ZPM prestatiekaart
        self.download_zpm_nza_excel(instelling)
        self.gebruiker.webscraper_hernoem_bestand(instelling, 'zpm_nza')

        # sluit het portaal af
        self.portaal.close()
        return None

    def webscraper_slm_portaal(self, instelling: Union[GGZ, ZKH]):
        # inloggen op het portaal
        self.inloggen(instelling)

        # download excel ZPM prestatiekaart
        rode_delta = self.get_slm_delta(instelling)
        instelling.set_slm_delta(rode_delta)

        # sluit het portaal af
        self.portaal.close()
        return None

    def webscraper_da_groot_portaal(self, instelling: Union[GGZ, ZKH]):
        # inloggen op het portaal
        self.inloggen(instelling)

        # download excel controle/norm in productie
        self.portaal.get(instelling.daily_audit)
        sleep(2)
        self.portaal.get(instelling.da_excel_groot_download)
        sleep(30)
        self.gebruiker.webscraper_hernoem_bestand(instelling, 'da', False)

        # sluit het portaal af
        self.portaal.close()
        return None

    # Webscraper Functies

    def webscraper_da(self, instelling: Union[GGZ, ZKH]):
        """Download de DA Excels van de opgegeven instelling (ZKH | GGZ)."""

        # download excels uit acceptatie omgeving
        instelling.kies_omgeving('acceptatie')
        self.webscraper_da_portaal(instelling)

        # download excels uit productie omgeving
        instelling.kies_omgeving('productie')
        self.webscraper_da_portaal(instelling)
        return None
    
    def webscraper_bi(self, instelling: Union[GGZ, ZKH]):
        """Download de BI Excels van de opgegeven instelling (ZKH | GGZ)."""

        # download excels uit acceptatie omgeving
        instelling.kies_omgeving('acceptatie')
        self.webscraper_bi_portaal(instelling)

        # download excels uit productie omgeving
        instelling.kies_omgeving('productie')
        self.webscraper_bi_portaal(instelling)
        return None

    def webscraper_zpm(self, instelling: Union[GGZ, ZKH]):
        """Download de ZPM Excels van de opgegeven instelling (ZKH | GGZ)."""

        # download excels uit acceptatie omgeving
        instelling.kies_omgeving('acceptatie')
        self.webscraper_zpm_portaal(instelling)

        # download excels uit productie omgeving
        instelling.kies_omgeving('productie')
        self.webscraper_zpm_portaal(instelling)
        return None
    
    def webscraper_zpm_nza(self, instelling: Union[GGZ, ZKH]):
        """Download de ZPM Excels van de opgegeven instelling (ZKH | GGZ)."""

        # download excels uit acceptatie omgeving
        instelling.kies_omgeving('acceptatie')
        self.webscraper_zpm_nza_portaal(instelling)

        # download excels uit productie omgeving
        instelling.kies_omgeving('productie')
        self.webscraper_zpm_nza_portaal(instelling)
        return None

    def webscraper_slm(self, instelling: Union[GGZ, ZKH]):
        """Download de BI Excels van de opgegeven instelling (ZKH | GGZ)."""

        # download excels uit acceptatie omgeving
        instelling.kies_omgeving('acceptatie')
        self.webscraper_slm_portaal(instelling)

        # download excels uit productie omgeving
        instelling.kies_omgeving('productie')
        self.webscraper_slm_portaal(instelling)
        return None

    def webscraper_da_groot(self, *instellingen: Union[GGZ, ZKH], omgeving: str = 'productie'):
        mislukt_download = []
        for instelling in instellingen:
            #try:
            #    instelling.kies_omgeving(omgeving)
            #    self.webscraper_da_groot_portaal(instelling)
            #except:
            #    mislukt_download.append(instelling.klant_code)
            #    print(f'Het downloaden van {instelling.klant_code} is mislukt!')
            instelling.kies_omgeving(omgeving)
            self.webscraper_da_groot_portaal(instelling)


        # Print mislukte downloads/tests
        if len(mislukt_download) != 0:
            print('Mislukte downloads:', ' '.join(mislukt_download))
        else:
            print('Geen mislukte downloads!')
        
        return None


# import Asclepius dependencies
from asclepius.instelling import GGZ, ZKH

# import other dependencies
import getpass
import os
import time
from typing import Union
import shutil

class Medewerker:
    
    def __init__(self, gebruikersnaam, downloads, bestemming):
        self.gebruikersnaam = gebruikersnaam
        self.wachtwoord = getpass.getpass(prompt='Wachtwoord: ')
        self.downloads = downloads
        self.bestemming = bestemming

    def laatste_download(self):
        # kijk welke bestanden er in downloads staan
        files = os.listdir(self.downloads)

        # converteer de gevonden namen naar paths
        paths = [os.path.join(self.downloads, basename) for basename in files]

        # vind het nieuwste bestand
        last_created_file = (max(paths, key = os.path.getctime))

        # vind de tijd waarop het nieuwste bestand is aangemaakt
        last = time.gmtime(os.path.getmtime(last_created_file))
        
        return last_created_file, last
    
    def genereer_nieuwe_map(self):
        today = time.gmtime()
        new_folder = f'{today.tm_year}-{today.tm_mon}-{today.tm_mday}'
        new_path = os.path.join(self.bestemming, new_folder)

        if not os.path.exists(new_path):
            os.makedirs(new_path)
        return new_path

    def webscraper_hernoem_bestand(self, instelling: Union[GGZ, ZKH], product: str, test: bool = False):
        last_created_file, last = self.laatste_download()

        today = time.gmtime()
        if last.tm_mday == today.tm_mday and last.tm_mon == today.tm_mon:
            # nieuwe bestandsnaam en nieuwe map
            new_name = instelling.genereer_nieuwe_naam(product, test)
            new_folder = self.genereer_nieuwe_map()
            new_path = os.path.join(new_folder, new_name)

            shutil.move(last_created_file, new_path)
            instelling.update_bestand_locatie(product, new_path, test)
        else:
            print("Bestand niet gevonden/hernoemen mislukt")
        return None



# import Asclepius dependencies
from pandas.core.frame import DataFrame
from asclepius.instelling import GGZ, ZKH
from asclepius.medewerker import Medewerker
from asclepius.portaaldriver import PortaalDriver
from asclepius.testen import TestFuncties, Verklaren

# import other dependencies
from typing import Union
from pandas import ExcelWriter

class ReleaseTesten:
    
    def __init__(self, gebruiker: Medewerker, losse_bestanden: bool = False):

        # Initialiseren
        self.gebruiker = gebruiker
        self.portaaldriver = PortaalDriver(self.gebruiker)
        self.testfuncties = TestFuncties()
        self.verklaren = Verklaren()

        self.losse_bestanden = losse_bestanden
        
        return None
    
    def test_da(self, *instellingen: Union[GGZ, ZKH]):
        # Download excelbestanden
        mislukt_download = []
        for instelling in instellingen:
            try:
                self.portaaldriver.webscraper_da(instelling)
            except:
                mislukt_download.append(instelling.klant_code)

        # Test de DA
        mislukt_da = []
        for instelling in instellingen:
            try:        
                # Aantallencheck
                self.testfuncties.aantallencheck(instelling, False)
                self.testfuncties.aantallencheck(instelling, True)

                # Standaardverschillen vinden
                self.verklaren.standaardverschillen_da(instelling, False)
                self.verklaren.standaardverschillen_da(instelling, True)
            except:
                mislukt_da.append(instelling.klant_code)
        
        if self.losse_bestanden:
            for instelling in instellingen:
                if instelling.klant_code not in set(mislukt_download + mislukt_da):
                    with ExcelWriter(f'Bevindingen DA {instelling.klant_code}.xlsx') as writer:
                        instelling.bevindingen_da.to_excel(writer, sheet_name=f'{instelling.klant_code}')
                        instelling.bevindingen_da_test.to_excel(writer, sheet_name=f'{instelling.klant_code} test')
                else: pass
        else:
            with ExcelWriter(f'Bevindingen DA.xlsx') as writer:
                for instelling in instellingen:
                    if instelling.klant_code not in set(mislukt_download + mislukt_da):
                        instelling.bevindingen_da.to_excel(writer, sheet_name=f'{instelling.klant_code}')
                        instelling.bevindingen_da_test.to_excel(writer, sheet_name=f'{instelling.klant_code} test')
                    else: pass

        # Print mislukte downloads/tests
        if len(mislukt_download) != 0:
            print('Mislukte downloads:', ' '.join(mislukt_download))
        else:
            print('Geen mislukte downloads!')
        
        if len(mislukt_da) != 0:
            print('Mislukte DA tests:', ' '.join(mislukt_da))
        else:
            print('Geen mislukte DA tests!')
        return None
    
    def test_bi(self, *instellingen: Union[GGZ, ZKH]):
        # Download excelbestanden
        mislukt_download = []
        for instelling in instellingen:
            try:
                self.portaaldriver.webscraper_bi(instelling)
            except:
                mislukt_download.append(instelling.klant_code)

        # Test de BI
        mislukt_bi = []
        for instelling in instellingen:
            try:        
                # Vergelijk BI prestatiekaarten
                self.testfuncties.prestatiekaarten_vergelijken(instelling, 'bi')
            except:
                mislukt_bi.append(instelling.klant_code)

        if self.losse_bestanden:
            for instelling in instellingen:
                if instelling.klant_code not in set(mislukt_download + mislukt_bi):
                    with ExcelWriter(f'Bevindingen BI {instelling.klant_code}.xlsx') as writer:
                        instelling.bevindingen_bi.to_excel(writer, sheet_name=f'{instelling.klant_code}')
                else: pass
        else:
            with ExcelWriter(f'Bevindingen BI.xlsx') as writer:
                for instelling in instellingen:
                    if instelling.klant_code not in set(mislukt_download + mislukt_bi):
                        instelling.bevindingen_bi.to_excel(writer, sheet_name=f'{instelling.klant_code}')
                    else: pass
        
        # Print mislukte downloads/tests
        if len(mislukt_download) != 0:
            print('Mislukte downloads:', ' '.join(mislukt_download))
        else:
            print('Geen mislukte downloads!')

        if len(mislukt_bi) != 0:
            print('Mislukte BI tests:', ' '.join(mislukt_bi))
        else:
            print('Geen mislukte BI tests!')
        return None
    
    def test_zpm(self, *instellingen: Union[GGZ, ZKH]):
        # Download excelbestanden
        mislukt_download = []
        for instelling in instellingen:
            try:
                self.portaaldriver.webscraper_zpm(instelling)
            except:
                mislukt_download.append(instelling.klant_code)

        # Test de ZPM
        mislukt_zpm = []
        for instelling in instellingen:
            try:        
                # Vergelijk ZPM prestatiekaarten
                self.testfuncties.prestatiekaarten_vergelijken(instelling, 'zpm')
            except:
                mislukt_zpm.append(instelling.klant_code)

        if self.losse_bestanden:
            for instelling in instellingen:
                if instelling.klant_code not in set(mislukt_download + mislukt_zpm):
                    with ExcelWriter(f'Bevindingen ZPM {instelling.klant_code}.xlsx') as writer:
                        instelling.bevindingen_zpm.to_excel(writer, sheet_name=f'{instelling.klant_code}')
                else: pass
        else:
            with ExcelWriter(f'Bevindingen ZPM.xlsx') as writer:
                for instelling in instellingen:
                    if instelling.klant_code not in set(mislukt_download + mislukt_zpm):
                        instelling.bevindingen_zpm.to_excel(writer, sheet_name=f'{instelling.klant_code}')
                    else: pass

        # Print mislukte downloads/tests
        if len(mislukt_download) != 0:
            print('Mislukte downloads:', ' '.join(mislukt_download))
        else:
            print('Geen mislukte downloads!')
        
        if len(mislukt_zpm) != 0:
            print('Mislukte ZPM tests:', ' '.join(mislukt_zpm))
        else:
            print('Geen mislukte ZPM tests!')
        return None

    def test_zpm_nza(self, *instellingen: Union[GGZ, ZKH]):
        # Download excelbestanden
        mislukt_download = []
        for instelling in instellingen:
            try:
                self.portaaldriver.webscraper_zpm_nza(instelling)
            except:
                mislukt_download.append(instelling.klant_code)

        # Test de ZPM_NZA
        mislukt_zpm = []
        for instelling in instellingen:
            try:        
                # Vergelijk ZPM_NZA prestatiekaarten
                self.testfuncties.prestatiekaarten_vergelijken(instelling, 'zpm_nza')
            except:
                mislukt_zpm.append(instelling.klant_code)

        if self.losse_bestanden:
            for instelling in instellingen:
                if instelling.klant_code not in set(mislukt_download + mislukt_zpm):
                    with ExcelWriter(f'Bevindingen ZPM 100% NZA {instelling.klant_code}.xlsx') as writer:
                        instelling.bevindingen_zpm_nza.to_excel(writer, sheet_name=f'{instelling.klant_code}')
                else: pass
        else:
            with ExcelWriter(f'Bevindingen ZPM 100% NZA.xlsx') as writer:
                for instelling in instellingen:
                    if instelling.klant_code not in set(mislukt_download + mislukt_zpm):
                        instelling.bevindingen_zpm_nza.to_excel(writer, sheet_name=f'{instelling.klant_code}')
                    else: pass

        # Print mislukte downloads/tests
        if len(mislukt_download) != 0:
            print('Mislukte downloads:', ' '.join(mislukt_download))
        else:
            print('Geen mislukte downloads!')
        
        if len(mislukt_zpm) != 0:
            print('Mislukte ZPM 100% NZA tests:', ' '.join(mislukt_zpm))
        else:
            print('Geen mislukte ZPM 100% NZA tests!')
        return None

    def test_slm(self, *instellingen: Union[GGZ, ZKH]):
        # Download excelbestanden
        mislukt_download = []
        for instelling in instellingen:
            try:
                self.portaaldriver.webscraper_slm(instelling)
            except:
                mislukt_download.append(instelling.klant_code)

        # Test de BI
        bevindingen_slm = DataFrame({'Instelling': [], 'Delta totaal A': [], 'Delta totaal P': []})
        for instelling in instellingen:
            if instelling.klant_code not in set(mislukt_download):
                    new_row = {'Instelling': instelling.klant_code, 'Delta totaal A': instelling.slm_delta_a, 'Delta totaal P': instelling.slm_delta_p}
                    bevindingen_slm = bevindingen_slm.append(new_row, ignore_index = True)
            else: pass

        with ExcelWriter(f'Bevindingen SLM.xlsx') as writer:
            for instelling in instellingen:
                if instelling.klant_code not in set(mislukt_download):
                    bevindingen_slm.to_excel(writer, sheet_name=f'{instelling.klant_code}')
                else: pass
        
        # Print mislukte downloads/tests
        if len(mislukt_download) != 0:
            print('Mislukte downloads:', ' '.join(mislukt_download))
        else:
            print('Geen mislukte downloads!')
        return None
    
